#=======================================================================
#	$Id: Triples.pm,v 1.8 2008/07/16 16:49:48 pythontech Exp $
#	Virtual base class for storage of triples
#	Copyright (C) 2004-2008  Python Technology Limited
#
#	This program is free software; you can redistribute it and/or
#	modify it under the terms of the GNU General Public License
#	as published by the Free Software Foundation; either version 2
#	of the License, or (at your option) any later version.
#
#	This program is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with this program; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  
#	02111-1307, USA.
#=======================================================================
package PythonTech::Triples;
use strict;

sub new {
    my $class = shift;
    die "Called virtual ${class}::new\n";
}

sub configure {
    my $self = shift;
    die "Called virtual ${self}->configure\n";
}

#-----------------------------------------------------------------------
#	Copy statements to another triple store
#-----------------------------------------------------------------------
sub copy {
    my($self, $other) = @_;
    $self->iterate(undef,undef,undef, sub {
	$other->add(@_);
	undef;
    });
}

#-----------------------------------------------------------------------
#	Import triples from another triplestore
#-----------------------------------------------------------------------
sub copy_in {
    my($self, $other) = @_;
    $other->iterate(undef,undef,undef, sub {
	$self->add(@_);
	undef;
    });
}

#-----------------------------------------------------------------------
#	Add a statement in the database
#-----------------------------------------------------------------------
sub add {
    my($self, $subj,$pred,$obj) = @_;
    die "Called virtual ${self}->add\n";
}

#-----------------------------------------------------------------------
#	Remove a statement from the database
#-----------------------------------------------------------------------
sub remove {
    my($self, $subj,$pred,$obj) = @_;
    die "Called virtual ${self}->remove\n";
}

#-----------------------------------------------------------------------
#	Iterate over matching statements, invoking a callback.
#	When callback returns a true value, exit loop and return it.
#-----------------------------------------------------------------------
sub iterate {
    my($self, $subj,$pred,$obj,$cb) = @_;
    die "Called virtual ${self}->iterate\n";
}

#-----------------------------------------------------------------------
#	Run a general query, returning a list of sublists (or only the
#	first, if any, in scalar context).  Each sublist contains the
#	unknowns of a matching statement.
#-----------------------------------------------------------------------
sub query {
    my($self, $subj,$pred,$obj) = @_;
    if (wantarray) {
	my @res;
	$self->iterate($subj,$pred,$obj, sub {
	    my @spo;
	    push @spo, $_[0] unless defined $subj;
	    push @spo, $_[1] unless defined $pred;
	    push @spo, $_[2] unless defined $obj;
	    push @res, \@spo;
	    undef;		# Keep looking
	});
	return @res;
    } else {
	my $res;
	$self->iterate($subj,$pred,$obj, sub {
	    my @spo;
	    push @spo, $_[0] unless defined $subj;
	    push @spo, $_[1] unless defined $pred;
	    push @spo, $_[2] unless defined $obj;
	    $res = \@spo;
	    1;			# Bale out
	});
	return $res;
    }
}

#-----------------------------------------------------------------------
#	Generic query
#-----------------------------------------------------------------------
sub query3 {
    my($self, $subj,$pred,$obj) = @_;
    if (wantarray) {
	my @res;		# List of lists
	$self->iterate($subj,$pred,$obj, sub {
	    push @res, [@_];
	    undef;		# Keep looking
	});
	return @res;
    } else {
	# Want only the first (if any)
	my $res;
	$self->iterate($subj,$pred,$obj, sub {
	    $res = [@_];
	    1;			# Bale out
	});
	return $res;
    }
}

#-----------------------------------------------------------------------
#	Test a statement
#-----------------------------------------------------------------------
sub test {
    my($self, $subj,$pred,$obj) = @_;
    return $self->iterate($subj,$pred,$obj, sub {1});
}

#-----------------------------------------------------------------------
#	Given subject and predicate, list objects
#-----------------------------------------------------------------------
sub sp2o {
    my($self, $subj,$pred) = @_;
    if (wantarray) {
	my @res;
	$self->iterate($subj,$pred,undef, sub {
	    push @res, $_[2];
	    undef;		# Keep looking
	});
	return @res;
    } else {
	my $res;
	$self->iterate($subj,$pred,undef, sub {
	    $res = $_[2];
	    1;			# Bale out
	});
	return $res;
    }
}

#-----------------------------------------------------------------------
#	Given predicate and object, list subjects
#-----------------------------------------------------------------------
sub po2s {
    my($self, $pred,$obj) = @_;
    if (wantarray) {
	my @res;
	$self->iterate(undef,$pred,$obj, sub {
	    push @res, $_[0];
	    undef;		# Keep looking
	});
	return @res;
    } else {
	my $res;
	$self->iterate(undef,$pred,$obj, sub {
	    $res = $_[0];
	    1;			# Bale out
	});
	return $res;
    }
}

#-----------------------------------------------------------------------
#	Given subject, list predicates
#-----------------------------------------------------------------------
sub s2p {
    my($self, $subj) = @_;
    if (wantarray) {
	my %pred;
	$self->iterate($subj,undef,undef, sub {
	    $pred{$_[1]} = 1;
	    undef;		# Keep looking
	});
	return keys %pred;
    } else {
	my $pred;
	$self->iterate($subj,undef,undef, sub {
	    $pred = $_[1];
	    1;			# Bale out
	});
	return $pred;
    }
}

#-----------------------------------------------------------------------
#	Given object, list predicates
#-----------------------------------------------------------------------
sub o2p {
    my($self, $obj) = @_;
    if (wantarray) {
	my %pred;
	$self->iterate(undef,undef,$obj, sub {
	    $pred{$_[1]} = 1;
	    undef;		# Keep looking
	});
	return keys %pred;
    } else {
	my $pred;
	$self->iterate(undef,undef,$obj, sub {
	    $pred = $_[1];
	    1;			# Bale out
	});
	return $pred;
    }
}

#-----------------------------------------------------------------------
#	Commit or back out of pending changes
#-----------------------------------------------------------------------
sub commit {
    die ref($_[0])." does not implement 'commit'\n";
}
sub rollback {
    die ref($_[0])." does not implement 'rollback'\n";
}

#-----------------------------------------------------------------------
#	Disconnect from / reconnect to persistence layer
#-----------------------------------------------------------------------
sub disconnect {
    die ref($_[0])." does not implement 'disconnect'\n";
}
sub reconnect {
    die ref($_[0])." does not implement 'reconnect'\n";
}

#-----------------------------------------------------------------------
#	Write as N-Triples
#-----------------------------------------------------------------------
sub write_ntriples {
    my($self, $filename) = @_;
    if (! defined $filename) {
	open(NT,'>&STDOUT') || die "Cannot open stdout: $!\n";
    } else {
	open(NT,'>',$filename) || die "Cannot open $filename: $!\n";
    }
    $self->iterate(undef,undef,undef, sub {
	my($subj,$pred,$obj) = @_;
	my $tobj;
	if (ord($obj) == ord('=')) {
	    $tobj = &literal_as_nt(substr($obj,1));
	} else {
	    $tobj = "<$obj>";
	}
	print NT "<$subj> <$pred> $tobj .\n";
	undef;			# Keep looking
    });
    close(NT);
}

#-----------------------------------------------------------------------
#	Format a URI reference or literal for use in N-Triples format
#	DEPRECATED: instance method, but instance irrelevant; provided
#	for Sverker
#-----------------------------------------------------------------------
sub format_object {
    my($self, $obj) = @_;
    my $nt;
    if ($obj =~ /^=/o) {
	$nt = &literal_as_nt(substr($obj,1));
    } else {
	$nt = &uriref_as_nt($obj);
    }
    return $nt;
}

#-----------------------------------------------------------------------
#	Parse a line containing an N-Triple
#-----------------------------------------------------------------------
sub parse_ntriple_line {
    my($self,$line) = @_;
    if ($line =~ /\S/) {
	if (my($subj,$pred,$ouri,$olit) =
	    $line =~ /^\s*
	    <([^\n>]+)>		# Subject as URI
	    \s+
	    <([^\n>]+)>		# Predicate as URI
	    \s+
	    (?:<([^\n>]*)>	# Object as URI
	     |
	     \"(
		(?:
		 \\.
		 |
		 [^\\\n\"]+
		 )*?
		)\"		# Object as escaped literal
	     (?:@\w+)?		# Ignore language tag
	     )
	    \s+
	    \.\s*\n?$/x) {
#	    if (0== (($subj) = $line =~ /^\s*<([^\n>]*)>/g)) {
#		die "Bad subject at line $lineno\n";
#	    } elsif (0== (($pred) = $line =~ /\G\s+<([^\n>]*)>/gc)) {
#		die "Bad predicate at line $lineno\n";
#	    } elsif (0== (($ourl,$olit) = $line =~ /\G\s+
#		die "Bad object at line $lineno\n";
#	    } else {
	    if (defined($ouri)) {
		$self->add($subj,$pred,$ouri);
	    } else {
		my $lit = &_unescape_literal($olit);
		$self->add($subj,$pred,"=$lit");
	    }
	} else {
	    die "N-Triple unrecognised: $line\n";
	}
    }
}

sub _unescape_literal {
    my($esc) = @_;
    my $lit = '';
    pos($esc) = 0;
    while (pos($esc) < length($esc)) {
	if ($esc =~ m!\G\\([nrt])!gcs) {
	    $lit .= {n=>"\n",r=>"\r",t=>"\t"}->{$1};
	} elsif ($esc =~ m!\G\\u([0-9a-fA-F]{4})!gcs) {
	    $lit .= chr(hex($1));
	} elsif ($esc =~ m!\G\\([\\\"])!gcs) {
	    $lit .= $1;
	} elsif ($esc =~ m!\G\\(.?)!gcs) {
	    die "Invalid escape $1 at position ".pos($esc);
	} elsif ($esc =~ m!\G([^\\]+)!gcs) {
	    $lit .= $1;
	} else {
	    die "Unexpected error in _unescape_literal";
	}
    }
    return $lit;
}

#-----------------------------------------------------------------------
#	Read ntriples from a file handle
#-----------------------------------------------------------------------
sub read_ntriples {
    my($self,$fh) = @_;
    my $lineno = 0;
    while (defined(my $line = <$fh>)) {
	++ $lineno;
	if (! eval {$self->parse_ntriple_line($line), 1}) {
	    chomp(my $err = $@);
	    die "$err at line $.\n";
	}
    }
}

#=======================================================================
#	Class methods
#=======================================================================

#-----------------------------------------------------------------------
#	Format a URI reference as one of the three items to be output
#	as an N-Triples statement
#-----------------------------------------------------------------------
sub uriref_as_nt {
    my $uri = shift;
    return "<$uri>";
}

#-----------------------------------------------------------------------
#	Format a literal as one of the three items to be output
#	as an N-Triples statement
#	FIXME charset considerations?
#	FIXME language considerations?
#-----------------------------------------------------------------------
sub literal_as_nt {
    my $lit = shift;
    # Escape backslash and quotes with backslash...
    $lit =~ s/([\\\"])/\\$1/g;
    # Newline as \n, return as \r, tab as \t
    $lit =~ s/\n/\\n/g;
    $lit =~ s/\r/\\r/g;
    $lit =~ s/\t/\\t/g;
    # Non-ASCII as \uXXXX
    $lit =~ s/([\x00-\x1f\x7f])/sprintf("\\u%04X",ord($1))/eg;
    return '"' . $lit . '"';
}

1;
